/**
 * Copyright (c) 2009-2011 Zauber S.A. <http://www.zaubersoftware.com/>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ar.com.zauber.commons.ws.connection;

import java.net.URL;
import java.util.Collection;
import java.util.LinkedList;
import java.util.Map;

import javax.xml.namespace.QName;
import javax.xml.ws.BindingProvider;
import javax.xml.ws.Service;
import javax.xml.ws.handler.soap.SOAPHandler;
import javax.xml.ws.handler.soap.SOAPMessageContext;

import ar.com.zauber.commons.ws.SoapHandlers;
import ar.com.zauber.commons.ws.security.UsernameTokenProfile;

/**
 * ConnectionBuilder
 *
 * @author flbulgarelli
 */
public class ConnectionBuilder<A> {

    private Class<? extends Service> serviceType;
    private Class<A> portType;
    private URL wsdlUrl;
    private String endpoint;
    private PortFactory portFactory = new Basic();
    private Collection<SOAPHandler<SOAPMessageContext>> headers = new LinkedList<SOAPHandler<SOAPMessageContext>>();
    private Collection<Configuration<? super A>> configurations = new LinkedList<Configuration<? super A>>();


    public ConnectionBuilder(Class<A> portType) {
        this.portType = portType;
    }

    /**mandatory*/
    public static <A> ConnectionBuilder<A> fromPortType(Class<A> portType) {
        return new ConnectionBuilder<A>(portType);
    }

    /**mandatory*/
    public ConnectionBuilder<A> withServiceType(Class<? extends Service> serviceType) {
        this.serviceType = serviceType;
        return this;
    }


    /**mandatory when {@link #withClasspathWsdl(String)} is not used*/
    public ConnectionBuilder<A> withWsdlUrl(URL wsdlUrl) {
        this.wsdlUrl = wsdlUrl;
        return this;
    }
    
    /**mandatory when {@link #withWsdlUrl(URL)} is not used*/
    public ConnectionBuilder<A> withClasspathWsdl(String resourceName) {
        URL resource = Thread.currentThread().getContextClassLoader().getResource(resourceName);
        if (resource == null) {
            throw new IllegalArgumentException("Resource " + resourceName + " not found");
        }
        return withWsdlUrl(resource);
    }

    /**optional*/
    public ConnectionBuilder<A> withEndpoint(String endpoint) {
        this.endpoint = endpoint;
        return this;
    }

    /**optional*/
    public ConnectionBuilder<A> withHeader(SOAPHandler<SOAPMessageContext> header) {
        this.headers.add(header);
        return this;
    }
    
    /**optional*/
    public ConnectionBuilder<A> withPortFactory(PortFactory portFactory) {
        this.portFactory = portFactory;
        return this;
    }
    
    /**optional*/
    public ConnectionBuilder<A> withPortQName(String qname) {
        return withPortQName(QName.valueOf(qname));
    }
    
    /**optional*/
    public ConnectionBuilder<A> withPortQName(QName qname) {
        return withPortFactory(new WithQName(qname));
    }

    /**optional*/
    public ConnectionBuilder<A> withUsernameTokenAuth(String username, String password) {
        return withHeader(new UsernameTokenProfile(username, password));
    }

    /**optional*/
    public ConnectionBuilder<A> withConfiguration(Configuration<? super A> configuration) {
        configurations.add(configuration);
        return this;
    }

    public A build() {
        try {
            if(wsdlUrl== null)
                throw new IllegalArgumentException("The wsdl Url or Classpath is missing");
            Object service = createService(serviceType, wsdlUrl);
            A port = portFactory.createPort(service, portType);
            configureEndpoint(port);
            configureHeaders(port);
            configureHooks(port);
            return port;
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    protected void configureHooks(A port) {
        for (Configuration<? super A> config : configurations)
            config.configure(port);
    }

    protected void configureHeaders(A port) {
        for (SOAPHandler<SOAPMessageContext> header : headers)
            SoapHandlers.add(provideBindings(port), header);
    }

    protected void configureEndpoint(A port) {
        if(endpoint != null) {
            Map<String, Object> requestContext = provideBindings(port).getRequestContext();
            requestContext.put(BindingProvider.ENDPOINT_ADDRESS_PROPERTY, endpoint);
        }
    }
    
   protected Object createService(Class<? extends Service> serviceType, URL wsdlUrl) throws Exception {
      return serviceType.getConstructor(URL.class).newInstance(wsdlUrl);
   }
    
    protected BindingProvider provideBindings(A port) {
       return (BindingProvider) port;
    }

    static class Basic implements PortFactory    {
      public <A> A createPort(Object service, Class<A> portType) {
         return ((Service)service).getPort(portType);
      }
    }
    
    static class WithQName implements PortFactory    {
      QName portName;
      public WithQName(QName portName) {
         this.portName = portName;
      }
      public <A> A createPort(Object service, Class<A> portType) {
         return ((Service)service).getPort(portName, portType);
      }
    }
}
